#include "CUDALut1D.h"

#include <cuda_runtime.h> 
#include <cutil.h>
#include <cutil_math.h>

texture<float4,1> TFTex;
__constant__ float c_std_sample_distance;

#ifndef __BasicTFParam__
#define __BasicTFParam__
__constant__ float c_tf_scale;
__constant__ float c_tf_shift;
#endif 


static __device__
float4 d_MapScalarToColor(float scalar)
{
	float Index=scalar*c_tf_scale+c_tf_shift;
	return tex1D(TFTex,Index);
}

static __device__
float4 d_MapScalarToColor(float scalar,float sampleDistance)
{
	float Index=scalar*c_tf_scale+c_tf_shift;
	float4 result=tex1D(TFTex,Index);
	result.w=result.w<1.0f?1.0f-powf(1.0f-result.w,sampleDistance/c_std_sample_distance):1.0f;
	return result;
}

static bool TF1DPrepare(const CUDALut1D *lut)
{
	if (!lut) return false;

	int sampleNum=lut->m_SampleNumber;
	float minPos=lut->m_min;
	float maxPos=lut->m_max;

	cudaChannelFormatDesc channelDesc;

	TFTex.normalized = false; 
    TFTex.filterMode = cudaFilterModeLinear;
    TFTex.addressMode[0] = cudaAddressModeClamp;  

	cudaArray* tfArray=lut->m_d_tfArray;

	CUDA_SAFE_CALL(cudaGetChannelDesc(&channelDesc, tfArray));
	CUDA_SAFE_CALL(cudaBindTextureToArray(TFTex,tfArray, channelDesc));

	float scale=(sampleNum-1)/(maxPos-minPos);
	float shift=-(sampleNum-1)*minPos/(maxPos-minPos)+0.5f;
	
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_tf_scale, &scale, sizeof(float),0,cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_tf_shift, &shift, sizeof(float),0,cudaMemcpyHostToDevice));

	return true;
}

static bool TF1DPrepare(const CUDALut1D *lut,float stdSampleDistance)
{
	if (! TF1DPrepare(lut)) return false;	

	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_std_sample_distance, &stdSampleDistance, sizeof(float),0,cudaMemcpyHostToDevice));
	
	return true;
}

static void TF1DClear()
{
	CUDA_SAFE_CALL(cudaUnbindTexture(TFTex));
}
