#include "CUDATFEmptySpace.h"

#include <cuda_runtime.h> 
#include <cutil.h>
#include <cutil_math.h>

texture<float,1> alphaAccumTex;

#ifndef __BasicTFParam__
#define __BasicTFParam__
__constant__ float c_tf_scale;
__constant__ float c_tf_shift;
#endif 

static __device__
bool d_EmptyTest(float minV,float maxV)
{
	float Index1=minV*c_tf_scale+c_tf_shift;
	float Index2=maxV*c_tf_scale+c_tf_shift;

	return tex1D(alphaAccumTex,Index2+0.5f)>tex1D(alphaAccumTex,Index1+0.5f); 
}

static __device__
float2 d_GetMinMax(int x,int y,int z);

static __device__
bool d_BlockEmptyTest(int x, int y, int z)
{
	float2 MinMax=d_GetMinMax(x,y,z);
	return d_EmptyTest(MinMax.x,MinMax.y);
}


static bool TFEmptySpacePrepare(const CUDATFEmptySpace *emptySpace)
{
	if (!emptySpace) return false;

	int sampleNum=emptySpace->m_SampleNumber;
	float minPos=emptySpace->m_min;
	float maxPos=emptySpace->m_max;

	cudaChannelFormatDesc channelDesc;

	alphaAccumTex.normalized = false; 
    alphaAccumTex.filterMode = cudaFilterModeLinear;
    alphaAccumTex.addressMode[0] = cudaAddressModeClamp;  

	cudaArray* accuArray=emptySpace->m_d_AlphaAccumArray;

	CUDA_SAFE_CALL(cudaGetChannelDesc(&channelDesc,accuArray));
	CUDA_SAFE_CALL(cudaBindTextureToArray(alphaAccumTex,accuArray, channelDesc));

	float scale=(sampleNum-1)/(maxPos-minPos);
	float shift=-(sampleNum-1)*minPos/(maxPos-minPos)+0.5f;
	
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_tf_scale, &scale, sizeof(float),0,cudaMemcpyHostToDevice));
	CUDA_SAFE_CALL(cudaMemcpyToSymbol(c_tf_shift, &shift, sizeof(float),0,cudaMemcpyHostToDevice));
	
	return true;
}

static void TFEmptySpaceClear()
{
	CUDA_SAFE_CALL(cudaUnbindTexture(alphaAccumTex));
}
