#include "CUDATypicalRayProducer_basic.cu.h"

static __device__ void d_Initialize();

static __global__ 
void g_GenerateImage()
{
	d_Initialize();
	unsigned x = blockIdx.x*blockDim.x + threadIdx.x;
	unsigned y = blockIdx.y*blockDim.y + threadIdx.y;

	if(x>=c_imageSize[0] || y>=c_imageSize[1]) return;

	float pixX,pixY;
	pixX = x + c_imageOrigin[0]+0.5f;
	pixY = y + c_imageOrigin[1]+0.5f;

	float4 voxel_zero, voxel_unit, CameraDirection;
	float tnear,tfar;
	d_ProduceRay(pixX,pixY,voxel_zero, voxel_unit, CameraDirection, tnear, tfar);

	float *pImage=c_d_image+(y*c_imageSize[0]+x)*4;

	float4 value=make_float4(0.0f);
	if (tnear<tfar)	value = d_CastRay(voxel_zero,voxel_unit,CameraDirection,tnear,tfar);

	pImage[0]=value.x;
	pImage[1]=value.y;
	pImage[2]=value.z;
	pImage[3]=value.w;	
	
}

static dim3 blockSize, gridSize;
static dim3 GetBlockSize()
{
	return blockSize;
}
static dim3 GetGridSize()
{
	return gridSize;
}

static bool GenerateImage(CUDAVRImage *image,const VRSceneParam* sceneParam, int cudaBlockSize)
{
	if (!TypicalRayProducerPrepare(image,sceneParam)) return false;

	int imageSize[2];
	imageSize[0]=image->m_Size[0];
	imageSize[1]=image->m_Size[1];
	
	blockSize.x = blockSize.y = cudaBlockSize;
	gridSize.x=iDivUp(imageSize[0], blockSize.x);
	gridSize.y=iDivUp(imageSize[1], blockSize.y);

	g_GenerateImage<<<gridSize, blockSize>>>();

	return true;
}

