#include "CUDAFullRayCaster_Preint_basic.cu.h"

static __device__ int ConvertToBlockIndex(int voxelIndex,int dimId);
static __device__ int ConvertToVoxelIndex(int blockIndex,int dimId);
static __device__ bool d_BlockEmptyTest(int x, int y, int z);

static __device__
float4 d_CastRay(const float4& voxel_zero,const float4& voxel_unit,const float4& CameraDirection,float tnear,float tfar)
{
	float t_b1=tnear;

	float4 pos_b1;
	pos_b1=voxel_zero+voxel_unit*t_b1;

	float tMaxX,tMaxY,tMaxZ;
	float tDeltaX,tDeltaY,tDeltaZ;

	float t_b2;
	
	float4 acc=make_float4(0.0f);
	float4 voxel_step=voxel_unit*c_sampleDistance;

	while (t_b1<tfar  && acc.w<0.98f)
	{
		{
			int x,y,z;
			int stepX,stepY,stepZ;

			if (voxel_unit.x==0)
				stepX=0;
			else if (voxel_unit.x<0)
				stepX=-1;
			else
				stepX=1;

			if (voxel_unit.y==0)
				stepY=0;
			else if (voxel_unit.y<0)
				stepY=-1;
			else
				stepY=1;

			if (voxel_unit.z==0)
				stepZ=0;
			else if (voxel_unit.z<0)
				stepZ=-1;
			else
				stepZ=1;
			pos_b1=voxel_zero+voxel_unit*t_b1;

			if (voxel_unit.x==0)
			{
				x=ConvertToBlockIndex(((int)ceilf(pos_b1.x-0.5f)-1),0);
				tMaxX=tfar;
			}
			else if (voxel_unit.x<0)
			{
				x=ConvertToBlockIndex(((int)ceilf(pos_b1.x-0.5f)-1),0);
				tMaxX=(ConvertToVoxelIndex(x,0)+0.5f-voxel_zero.x)/voxel_unit.x;
				tDeltaX=-ConvertToVoxelIndex(1,0)/voxel_unit.x;
			}
			else
			{
				x=ConvertToBlockIndex((int)floorf(pos_b1.x-0.5f),0);
				tMaxX=(ConvertToVoxelIndex(x+1,0)+0.5f-voxel_zero.x)/voxel_unit.x;
				tDeltaX=ConvertToVoxelIndex(1,0)/voxel_unit.x;
			}

			if (voxel_unit.y==0)
			{
				y=ConvertToBlockIndex(((int)ceilf(pos_b1.y-0.5f)-1),1);
				tMaxY=tfar;
			}
			else if (voxel_unit.y<0)
			{
				y=ConvertToBlockIndex(((int)ceilf(pos_b1.y-0.5f)-1),1);
				tMaxY=(ConvertToVoxelIndex(y,1)+0.5f-voxel_zero.y)/voxel_unit.y;
				tDeltaY=-ConvertToVoxelIndex(1,1)/voxel_unit.y;
			}
			else
			{
				y=ConvertToBlockIndex((int)floorf(pos_b1.y-0.5f),1);
				tMaxY=(ConvertToVoxelIndex(y+1,1)+0.5f-voxel_zero.y)/voxel_unit.y;
				tDeltaY=ConvertToVoxelIndex(1,1)/voxel_unit.y;
			}
			
			if (voxel_unit.z==0)
			{
				z=ConvertToBlockIndex(((int)ceilf(pos_b1.z-0.5f)-1),2);
				tMaxZ=tfar;
			}
			else if (voxel_unit.z<0)
			{
				z=ConvertToBlockIndex(((int)ceilf(pos_b1.z-0.5f)-1),2);
				tMaxZ=(ConvertToVoxelIndex(z,2)+0.5f-voxel_zero.z)/voxel_unit.z;
				tDeltaZ=-ConvertToVoxelIndex(1,2)/voxel_unit.z;
			}
			else
			{
				z=ConvertToBlockIndex((int)floorf(pos_b1.z-0.5f),2);
				tMaxZ=(ConvertToVoxelIndex(z+1,2)+0.5f-voxel_zero.z)/voxel_unit.z;
				tDeltaZ=ConvertToVoxelIndex(1,2)/voxel_unit.z;
			}

			while (t_b1<tfar)
			{
				t_b2=min(min(tMaxX,tMaxY),tMaxZ);
			
				if (t_b2>t_b1+0.001f && d_BlockEmptyTest(x,y,z))	
					break; 

				if (tMaxX<tMaxY)
				{
					if (tMaxX<tMaxZ) 
					{
						tMaxX+=tDeltaX;
						x+=stepX;
					}
					else 
					{
						tMaxZ+=tDeltaZ;
						z+=stepZ;
					}
				}
				else
				{
					if (tMaxY<tMaxZ) 
					{
						tMaxY+=tDeltaY;
						y+=stepY;
					}
					else 
					{
						tMaxZ+=tDeltaZ;
						z+=stepZ;
					}
				}
				t_b1=t_b2;
			}
			if (t_b1>=tfar) break;
			if (t_b2>tfar) t_b2=tfar;
		}

		float t0=t_b1;
		float4 pos0=voxel_zero + voxel_unit * t0;
		float4 voxel_step=voxel_unit*c_sampleDistance;

		float scalar0;
		float3 k0,b0;

		scalar0=d_GetScalarValue(pos0);
		d_VolumeModel_Shade(pos0,CameraDirection,k0,b0);

		float t1 = t0 + c_sampleDistance;
		float4 pos1 = pos0 +  voxel_step;

		float scalar1;
		float3 k1,b1;

		while(t0<t_b2 && acc.w<0.98f)
		{
			scalar1=d_GetScalarValue(pos1);
			d_VolumeModel_Shade(pos1,CameraDirection,k1,b1);

			float4 col1,col2;
		
			col1=tex2D(TFTex1,scalar1*c_tf_scale+c_tf_shift,scalar0*c_tf_scale+c_tf_shift);
			col2=tex2D(TFTex2,scalar1*c_tf_scale+c_tf_shift,scalar0*c_tf_scale+c_tf_shift);

			if (c_shade)
			{
				col1 = make_float4(make_float3(col1)*k0+b0*col1.w,col1.w);
				col2 = make_float4(make_float3(col2)*k1+b1*col2.w,col2.w);
			}
			
			acc+=(1.0f-acc.w)*(col1+col2);

			t0 = t1;
			pos0 = pos1;
			scalar0 = scalar1;
			k0=k1;
			b0=b1;

			t1 = t0 + c_sampleDistance;
			pos1 = pos0 +  voxel_step;
		}

		t_b1=t_b2;

	}	

	return acc;
}
