#include "CUDAFullRayCaster_basic.cu.h"

static __device__ int ConvertToBlockIndex(int voxelIndex,int dimId);
static __device__ int ConvertToVoxelIndex(int blockIndex,int dimId);
static __device__ bool d_BlockEmptyTest(int x, int y, int z);

static __device__
float4 d_CastRay(const float4& voxel_zero,const float4& voxel_unit,const float4& CameraDirection,float tnear,float tfar)
{
	float t1=tnear;

	float4 pos1;
	pos1=voxel_zero+voxel_unit*t1;

	float tMaxX,tMaxY,tMaxZ;
	float tDeltaX,tDeltaY,tDeltaZ;

	float t_b;
	
	float4 acc=make_float4(0.0f);
	float4 voxel_step=voxel_unit*c_sampleDistance;

	float remain=0.0f;

	while (t1<tfar  && acc.w<0.98f)
	{
		{
			int x,y,z;
			int stepX,stepY,stepZ;

			if (voxel_unit.x==0)
				stepX=0;
			else if (voxel_unit.x<0)
				stepX=-1;
			else
				stepX=1;

			if (voxel_unit.y==0)
				stepY=0;
			else if (voxel_unit.y<0)
				stepY=-1;
			else
				stepY=1;

			if (voxel_unit.z==0)
				stepZ=0;
			else if (voxel_unit.z<0)
				stepZ=-1;
			else
				stepZ=1;
			pos1=voxel_zero+voxel_unit*t1;

			if (voxel_unit.x==0)
			{
				x=ConvertToBlockIndex(((int)ceilf(pos1.x-0.5f)-1),0);
				tMaxX=tfar;
			}
			else if (voxel_unit.x<0)
			{
				x=ConvertToBlockIndex(((int)ceilf(pos1.x-0.5f)-1),0);
				tMaxX=(ConvertToVoxelIndex(x,0)+0.5f-voxel_zero.x)/voxel_unit.x;
				tDeltaX=-ConvertToVoxelIndex(1,0)/voxel_unit.x;
			}
			else
			{
				x=ConvertToBlockIndex((int)floorf(pos1.x-0.5f),0);
				tMaxX=(ConvertToVoxelIndex(x+1,0)+0.5f-voxel_zero.x)/voxel_unit.x;
				tDeltaX=ConvertToVoxelIndex(1,0)/voxel_unit.x;
			}

			if (voxel_unit.y==0)
			{
				y=ConvertToBlockIndex(((int)ceilf(pos1.y-0.5f)-1),1);
				tMaxY=tfar;
			}
			else if (voxel_unit.y<0)
			{
				y=ConvertToBlockIndex(((int)ceilf(pos1.y-0.5f)-1),1);
				tMaxY=(ConvertToVoxelIndex(y,1)+0.5f-voxel_zero.y)/voxel_unit.y;
				tDeltaY=-ConvertToVoxelIndex(1,1)/voxel_unit.y;
			}
			else
			{
				y=ConvertToBlockIndex((int)floorf(pos1.y-0.5f),1);
				tMaxY=(ConvertToVoxelIndex(y+1,1)+0.5f-voxel_zero.y)/voxel_unit.y;
				tDeltaY=ConvertToVoxelIndex(1,1)/voxel_unit.y;
			}
			
			if (voxel_unit.z==0)
			{
				z=ConvertToBlockIndex(((int)ceilf(pos1.z-0.5f)-1),2);
				tMaxZ=tfar;
			}
			else if (voxel_unit.z<0)
			{
				z=ConvertToBlockIndex(((int)ceilf(pos1.z-0.5f)-1),2);
				tMaxZ=(ConvertToVoxelIndex(z,2)+0.5f-voxel_zero.z)/voxel_unit.z;
				tDeltaZ=-ConvertToVoxelIndex(1,2)/voxel_unit.z;
			}
			else
			{
				z=ConvertToBlockIndex((int)floorf(pos1.z-0.5f),2);
				tMaxZ=(ConvertToVoxelIndex(z+1,2)+0.5f-voxel_zero.z)/voxel_unit.z;
				tDeltaZ=ConvertToVoxelIndex(1,2)/voxel_unit.z;
			}

			while (t1<tfar)
			{
				t_b=min(min(tMaxX,tMaxY),tMaxZ);
			
				if (t_b>t1+0.001f && d_BlockEmptyTest(x,y,z))	
					break; 

				if (tMaxX<tMaxY)
				{
					if (tMaxX<tMaxZ) 
					{
						tMaxX+=tDeltaX;
						x+=stepX;
					}
					else 
					{
						tMaxZ+=tDeltaZ;
						z+=stepZ;
					}
				}
				else
				{
					if (tMaxY<tMaxZ) 
					{
						tMaxY+=tDeltaY;
						y+=stepY;
					}
					else 
					{
						tMaxZ+=tDeltaZ;
						z+=stepZ;
					}
				}
				t1=t_b;
			}
			if (t1>=tfar) break;
			if (t_b>tfar) t_b=tfar;
		}

		float t=t1+remain;
		float4 pos=voxel_zero + voxel_unit * t;		

		while(t<t_b && acc.w<0.98f)
		{
			float4 color=d_GetShadedColor(pos,CameraDirection);
			float alpha=color.w;
			color.w=1.0f;
			color*=alpha;		
			acc+=(1.0f-acc.w)*color;				
			t += c_sampleDistance;
			pos += voxel_step;
		}

		remain=t-t_b;

		t1=t_b;

	}	

	return acc;
}
