#include "CUDAIsosurfaceRayCaster_basic.cu.h"

static __device__
float4 d_CastRay(const float4& voxel_zero,const float4& voxel_unit,const float4& CameraDirection,float tnear,float tfar)
{
	float t1=tnear;

	float4 pos1,pos4;
	pos1=voxel_zero+voxel_unit*t1;

	float tMaxX,tMaxY,tMaxZ;
	float tDeltaX,tDeltaY,tDeltaZ;

	if (voxel_unit.x==0)
	{
		tMaxX=tfar;
	}
	else if (voxel_unit.x<0)
	{
		float x=ceilf(pos1.x-0.5f)-0.5f;
		tMaxX=(x-voxel_zero.x)/voxel_unit.x;
		tDeltaX=-1.0f/voxel_unit.x;
	}
	else
	{
		float x=floorf(pos1.x-0.5f)+0.5f;
		tMaxX=(x+1.0f-voxel_zero.x)/voxel_unit.x;
		tDeltaX=1.0f/voxel_unit.x;
	}

	if (voxel_unit.y==0)
	{
		tMaxY=tfar;
	}
	else if (voxel_unit.y<0)
	{
		float y=ceilf(pos1.y-0.5f)-0.5f;
		tMaxY=(y-voxel_zero.y)/voxel_unit.y;
		tDeltaY=-1.0f/voxel_unit.y;
	}
	else
	{
		float y=floorf(pos1.y-0.5f)+0.5f;
		tMaxY=(y+1.0f-voxel_zero.y)/voxel_unit.y;
		tDeltaY=1.0f/voxel_unit.y;
	}

	if (voxel_unit.z==0)
	{
		tMaxZ=tfar;
	}
	else if (voxel_unit.z<0)
	{
		float z=ceilf(pos1.z-0.5f)-0.5f;
		tMaxZ=(z-voxel_zero.z)/voxel_unit.z;
		tDeltaZ=-1.0f/voxel_unit.z;
	}
	else
	{
		float z=floorf(pos1.z-0.5f)+0.5f;
		tMaxZ=(z+1.0f-voxel_zero.z)/voxel_unit.z;
		tDeltaZ=1.0f/voxel_unit.z;
	}

	
	float4 samples;
	samples.x=d_GetScalarValue(pos1);
	bool increasing=samples.x<c_isovalue;

	while(t1<tfar)
	{
		float midVert1;
		float midVert2;
		float t2,t3,t4;

		while (t1<tfar)
		{
			t4=min(min(tMaxX,tMaxY),tMaxZ);
			if (t4>tfar) t4=tfar;

			if (tMaxX<tMaxY)
			{
				if (tMaxX<tMaxZ) tMaxX+=tDeltaX;
				else tMaxZ+=tDeltaZ;
			}
			else
			{
				if (tMaxY<tMaxZ) tMaxY+=tDeltaY;
				else tMaxZ+=tDeltaZ;
			}

			pos4=voxel_zero+voxel_unit*t4;	
			samples.w=d_GetScalarValue(pos4);

			t2=(t4-t1)*(1.0f/3.0f)+t1;
			samples.y=d_GetScalarValue(voxel_zero+voxel_unit*t2);

			t3=(t4-t1)*(2.0f/3.0f)+t1;		
			samples.z=d_GetScalarValue(voxel_zero+voxel_unit*t3);

			midVert1=-(5.0f/6.0f)*samples.x+3.0f*samples.y-1.5*samples.z+(1.0f/3.0f)*samples.w;
			midVert2=(1.0f/3.0f)*samples.x-1.5*samples.y+3.0f*samples.z-(5.0f/6.0f)*samples.w;

			if (min(min(min(samples.x,samples.w),midVert1),midVert2)<c_isovalue && 
				max(max(max(samples.x,samples.w),midVert1),midVert2)>c_isovalue) break;

			t1=t4;
			pos1=pos4;
			samples.x=samples.w;
			
			
		}
		if (t1>=tfar) break;

		midVert1=d_GetScalarValue(make_float4(pos4.x,pos1.y,pos1.z,1.0f));
		midVert1+=d_GetScalarValue(make_float4(pos1.x,pos4.y,pos1.z,1.0f));
		midVert1+=d_GetScalarValue(make_float4(pos1.x,pos1.y,pos4.z,1.0f));

		midVert2=d_GetScalarValue(make_float4(pos1.x,pos4.y,pos4.z,1.0f));
		midVert2+=d_GetScalarValue(make_float4(pos4.x,pos1.y,pos4.z,1.0f));
		midVert2+=d_GetScalarValue(make_float4(pos4.x,pos4.y,pos1.z,1.0f));

		float a=-samples.x+midVert1-midVert2+samples.w;
		float b=3.0f*samples.x-2.0f*midVert1+midVert2;
		float c=-3.0f*samples.x+midVert1;
		float d=samples.x-c_isovalue;
		
		float root;
		
		if (d_Solve_Cubic_Eq(a,b,c,d,root))
		{
			float t=(t4-t1)*root+t1;
			return d_GetColor(voxel_zero+voxel_unit*t,CameraDirection,increasing);
		}

		t1=t4;
		pos1=pos4;
		samples.x=samples.w;

	}

	return make_float4(0.0f);
	
}



