#include "MIP/CUDAMIPRayCaster_basic.cu.h"

static __device__ int ConvertToBlockIndex(int voxelIndex,int dimId);
static __device__ int ConvertToVoxelIndex(int blockIndex,int dimId);
static __device__ float2 d_GetMinMax(int x,int y,int z);

static __device__
float4 d_CastRay(const float4& voxel_zero,const float4& voxel_unit,const float4& CameraDirection,float tnear,float tfar)
{
	float t1=tnear;

	float4 pos1,pos4;
	pos1=voxel_zero+voxel_unit*t1;

	float tMaxX_b,tMaxY_b,tMaxZ_b;
	float tDeltaX_b,tDeltaY_b,tDeltaZ_b;
	float tDeltaX,tDeltaY,tDeltaZ;

	int x_b,y_b,z_b;
	int stepX,stepY,stepZ;

	if (voxel_unit.x==0)
	{
		x_b=ConvertToBlockIndex(((int)ceilf(pos1.x-0.5f)-1),0);
		stepX=0;
		tMaxX_b=tfar;
	}
	else if (voxel_unit.x<0)
	{
		x_b=ConvertToBlockIndex(((int)ceilf(pos1.x-0.5f)-1),0);
		stepX=-1;
		tMaxX_b=(ConvertToVoxelIndex(x_b,0)+0.5f-voxel_zero.x)/voxel_unit.x;
		tDeltaX_b=-ConvertToVoxelIndex(1,0)/voxel_unit.x;
		tDeltaX=-1.0f/voxel_unit.x;
	}
	else
	{
		x_b=ConvertToBlockIndex((int)floorf(pos1.x-0.5f),0);
		stepX=1;
		tMaxX_b=(ConvertToVoxelIndex(x_b+1,0)+0.5f-voxel_zero.x)/voxel_unit.x;
		tDeltaX_b=ConvertToVoxelIndex(1,0)/voxel_unit.x;
		tDeltaX=1.0f/voxel_unit.x;
	}

	if (voxel_unit.y==0)
	{
		y_b=ConvertToBlockIndex(((int)ceilf(pos1.y-0.5f)-1),1);
		stepY=0;
		tMaxY_b=tfar;
	}
	else if (voxel_unit.y<0)
	{
		y_b=ConvertToBlockIndex(((int)ceilf(pos1.y-0.5f)-1),1);
		stepY=-1;
		tMaxY_b=(ConvertToVoxelIndex(y_b,1)+0.5f-voxel_zero.y)/voxel_unit.y;
		tDeltaY_b=-ConvertToVoxelIndex(1,1)/voxel_unit.y;
		tDeltaY=-1.0f/voxel_unit.y;
	}
	else
	{
		y_b=ConvertToBlockIndex((int)floorf(pos1.y-0.5f),1);
		stepY=1;
		tMaxY_b=(ConvertToVoxelIndex(y_b+1,1)+0.5f-voxel_zero.y)/voxel_unit.y;
		tDeltaY_b=ConvertToVoxelIndex(1,1)/voxel_unit.y;
		tDeltaY=1.0f/voxel_unit.y;
	}

	if (voxel_unit.z==0)
	{
		z_b=ConvertToBlockIndex(((int)ceilf(pos1.z-0.5f)-1),2);
		stepZ=0;
		tMaxZ_b=tfar;
	}
	else if (voxel_unit.z<0)
	{
		z_b=ConvertToBlockIndex(((int)ceilf(pos1.z-0.5f)-1),2);
		stepZ=-1;
		tMaxZ_b=(ConvertToVoxelIndex(z_b,2)+0.5f-voxel_zero.z)/voxel_unit.z;
		tDeltaZ_b=-ConvertToVoxelIndex(1,2)/voxel_unit.z;
		tDeltaZ=-1.0f/voxel_unit.z;
	}
	else
	{
		z_b=ConvertToBlockIndex((int)floorf(pos1.z-0.5f),2);
		stepZ=1;
		tMaxZ_b=(ConvertToVoxelIndex(z_b+1,2)+0.5f-voxel_zero.z)/voxel_unit.z;
		tDeltaZ_b=ConvertToVoxelIndex(1,2)/voxel_unit.z;
		tDeltaZ=1.0f/voxel_unit.z;
	}

	float4 samples;
	samples.x=d_GetScalarValue(pos1);
	float maxI=samples.x;

	float t_b;
	while (t1<tfar)
	{
		float2 MinMax=d_GetMinMax(x_b,y_b,z_b);
		t_b=min(min(tMaxX_b,tMaxY_b),tMaxZ_b);
		if (tMaxX_b<tMaxY_b)
		{
			if (tMaxX_b<tMaxZ_b) 
			{
				tMaxX_b+=tDeltaX_b;
				x_b+=stepX;
			}
			else 
			{
				tMaxZ_b+=tDeltaZ_b;
				z_b+=stepZ;
			}
		}
		else
		{
			if (tMaxY_b<tMaxZ_b) 
			{
				tMaxY_b+=tDeltaY_b;
				y_b+=stepY;
			}
			else 
			{
				tMaxZ_b+=tDeltaZ_b;
				z_b+=stepZ;
			}
		}
	
		if ( MinMax.y>maxI)
		{
			if (t_b>tfar) t_b=tfar;
			pos1=voxel_zero+voxel_unit*t1;
			
			float tMaxX,tMaxY,tMaxZ;
			

			if (voxel_unit.x==0)
			{
				tMaxX=tfar;
			}
			else if (voxel_unit.x<0)
			{
				float x=ceilf(pos1.x-0.5f)-0.5f;
				tMaxX=(x-voxel_zero.x)/voxel_unit.x;
			}
			else
			{
				float x=floorf(pos1.x-0.5f)+0.5f;
				tMaxX=(x+1.0f-voxel_zero.x)/voxel_unit.x;
			}

			if (voxel_unit.y==0)
			{
				tMaxY=tfar;
			}
			else if (voxel_unit.y<0)
			{
				float y=ceilf(pos1.y-0.5f)-0.5f;
				tMaxY=(y-voxel_zero.y)/voxel_unit.y;
			}
			else
			{
				float y=floorf(pos1.y-0.5f)+0.5f;
				tMaxY=(y+1.0f-voxel_zero.y)/voxel_unit.y;
			}

			if (voxel_unit.z==0)
			{
				tMaxZ=tfar;
			}
			else if (voxel_unit.z<0)
			{
				float z=ceilf(pos1.z-0.5f)-0.5f;
				tMaxZ=(z-voxel_zero.z)/voxel_unit.z;
			}
			else
			{
				float z=floorf(pos1.z-0.5f)+0.5f;
				tMaxZ=(z+1.0f-voxel_zero.z)/voxel_unit.z;
			}

			samples.x=d_GetScalarValue(pos1);			

			float midVert1;
			float midVert2;
			float t2,t3,t4;
			
			while (t1<t_b)
			{
				t4=min(min(tMaxX,tMaxY),tMaxZ);
				if (t4>t_b) t4=t_b;

				if (tMaxX<tMaxY)
				{
					if (tMaxX<tMaxZ) tMaxX+=tDeltaX;
					else tMaxZ+=tDeltaZ;
				}
				else
				{
					if (tMaxY<tMaxZ) tMaxY+=tDeltaY;
					else tMaxZ+=tDeltaZ;
				}

				pos4=voxel_zero+voxel_unit*t4;	
				samples.w=d_GetScalarValue(pos4);

				t2=(t4-t1)*(1.0f/3.0f)+t1;
				samples.y=d_GetScalarValue(voxel_zero+voxel_unit*t2);

				t3=(t4-t1)*(2.0f/3.0f)+t1;		
				samples.z=d_GetScalarValue(voxel_zero+voxel_unit*t3);

				midVert1=-(5.0f/6.0f)*samples.x+3.0f*samples.y-1.5*samples.z+(1.0f/3.0f)*samples.w;
				midVert2=(1.0f/3.0f)*samples.x-1.5*samples.y+3.0f*samples.z-(5.0f/6.0f)*samples.w;

				float _maxI=max(samples.x,max(samples.w,max(midVert1,midVert2)));
				if (_maxI>maxI)
				{
					if (_maxI==samples.x || _maxI==samples.w) maxI=_maxI;
					else 
					{
						float a=-samples.x+3.0f*midVert1-3.0f*midVert2+samples.w;
						float b=3.0f*samples.x-6.0f*midVert1+3.0f*midVert2;
						float c=-3.0f*samples.x+3.0f*midVert1;
						float d=samples.x;

						maxI=max(maxI,d_GetMax(a,b,c,d));
					}
				}

				t1=t4;
				pos1=pos4;
				samples.x=samples.w;

			}

		}

		t1=t_b;
		
	}
		
	maxI=maxI*c_mip_scale+c_mip_shift;
	return make_float4(maxI,maxI,maxI,1.0f);
}


