#ifndef __Scene_h
#define __Scene_h

#include "AbstractScene.h"
#include "SceneParam.h"

class VOLTK_SCENE_API Scene : public AbstractScene
{
public:
	Scene();
	virtual ~Scene();

	void SetFovy(float fovy);

	void SetRotateRadPerPixel(float radPerPixel) { m_RadPerPixel=radPerPixel;}
	float GetRotateRadPerPixel() { return m_RadPerPixel; }

	void SetScaleRatePerPixel(float scaleRatePerPixel) { m_ScaleRatePerPixel=scaleRatePerPixel;}
	float GetScaleRatePerPixel() { return m_ScaleRatePerPixel; }

	void GetTranslate(float translate[3]);

	float GetScale() { return m_Param.m_Scale;}
	const Quaternion& GetRotation() { return m_Param.m_Rotation; }

	void SetTranslate(const float translate[3]);
	void SetScale(float scale);

	void SetRotation(const Quaternion& rotation);
	const Matrix4x4& GetViewMatrix();
	const Matrix4x4& GetProjectionMatrix();

	float GetZTranslateBase() { return m_Param.m_ZTranslateBase;} 

	void GetViewPort(int* viewport) const { viewport[0]=m_ViewPort[0]; viewport[1]=m_ViewPort[1];}

	float GetFovy() {return m_Param.m_Fovy;}

	virtual void Render();
	virtual void Resize(int width, int height);

	virtual void Init();
	
	virtual void Keyboard ( unsigned char key, int x, int y, retStatus &status);
	virtual void KeyboardUp ( unsigned char key, int x, int y, retStatus &status);
	
	virtual void LButtonDown(int x, int y, retStatus &status);
	virtual void LButtonUp(int x, int y, retStatus &status);

	virtual void MButtonDown(int x, int y, retStatus &status);
	virtual void MButtonUp(int x, int y, retStatus &status);

	virtual void RButtonDown(int x, int y, retStatus &status);
	virtual void RButtonUp(int x, int y, retStatus &status);

	virtual void MouseMove(int x, int y, retStatus &status);

	void SetEndoscopy(bool turnOn);
	bool GetEndoscopy() { return m_Endoscopy;}

	void SetMoveSensitivity(float maxSpeed, float acceleartion);
	void GetMoveSensitivity(float& maxSpeed, float& acceleartion);

	void MoveTo(float x,float y, float z); 
	void Move(float dx, float dy, float dz); 
	void Turn(float dx, float dy);
	void Roll(float rad);

	SceneParam& GetSceneParameters() { return m_Param;}	


protected:
	virtual void RenderContent();
	virtual void GetSize3D(float size3D[3]);
	void SetSize3DModified();

	int m_OldMouseX,m_OldMouseY;


private:
// View Port
	int m_ViewPort[2];
	SceneParam m_Param;	
	
// Status
	bool m_NeedCalculateCamera;

// Matrix
	Matrix4x4 m_ViewMatrix;
	Matrix4x4 m_ProjectionMatrix;

////// Manipulation

// Manipulation Sensitivity
	float m_RadPerPixel;
	float m_ScaleRatePerPixel;

// Manipulation Status
	bool m_Rotating;
	bool m_Scaling;
	bool m_Moving;

	bool m_Endoscopy;
	bool m_EndoscopyDragging;
	bool m_EndoscopyDragging2;

	float m_MaxSpeed;
	float m_MoveSpeed;
	float m_Acceleration;

};

#endif
