#include "BasicLightParam.h"
#include "Vector4.h"
#include "Vector3.h"

static inline Vector4 GetColor(Vector4 pos,float sampleDistance);
static inline Vector4 GetNormal(const Vector4& pos);

static float sampleDistance;
static bool shade;
static Vector3 ambient;
static Vector3 diffuse;
static Vector3 specular;
static float specular_power;
static Vector4 light_direction;
static float spec_threshold;

static bool illustrative;
static Vector3 backgroundColor;

#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

#ifndef min
#define min(a,b)            (((a) < (b)) ? (a) : (b))
#endif

static inline Vector4 GetShadedColor(const Vector4& pos,const Vector4& CameraDirection)
{
	Vector4 Color=GetColor(pos,sampleDistance);
	if (shade)
	{
		Vector4 normal=GetNormal(pos);
		float n_dot_c;
		n_dot_c=normal*CameraDirection;
		
		if (illustrative)
		{
			float rate=min(powf(1.0f-n_dot_c*n_dot_c,specular_power*0.5f),1.0f);
			Color[0]=Color[0]*rate+backgroundColor[0]*(1.0f-rate);
			Color[1]=Color[1]*rate+backgroundColor[1]*(1.0f-rate);
			Color[2]=Color[2]*rate+backgroundColor[2]*(1.0f-rate);
		}
		else
		{
			Vector3 k=ambient;
			Vector3	b;

			float n_dot_l;
			float n_dot_h; 

			n_dot_l=normal*light_direction;
			
			if ((n_dot_l>0.0f && n_dot_c>0.0f) || (n_dot_l<0.0f && n_dot_c<0.0f))
			{
				n_dot_l=n_dot_l>0?n_dot_l:-n_dot_l;
				k+=diffuse*n_dot_l;
				Vector4 HalfwayVector=light_direction+CameraDirection;
				HalfwayVector.Normalize();
				n_dot_h=normal*HalfwayVector;
				if (n_dot_h<0) n_dot_h=-n_dot_h;
				if (n_dot_h > spec_threshold)
				{
					b=specular*powf(n_dot_h,specular_power);
				}
			}
			
			Color[0]=Color[0]*k[0]+b[0];
			Color[1]=Color[1]*k[1]+b[1];
			Color[2]=Color[2]*k[2]+b[2];
		}
	}

	return Color;
}

static bool FullRayCasterPrepare( bool b_shade, float SampleDistance, const BasicLightParam* param)
{
	if (!param) return false;

	::sampleDistance=SampleDistance;
	::shade=b_shade?1:0;
	
	float lightColor[3];
	lightColor[0]=param->m_LightColor[0];
	lightColor[1]=param->m_LightColor[1];
	lightColor[2]=param->m_LightColor[2];

	float LightIntensity=param->m_LightIntensity;

	lightColor[0]*=LightIntensity;
	lightColor[1]*=LightIntensity;
	lightColor[2]*=LightIntensity;

	float f_ambient=param->m_Ambient;
	::ambient[0]=f_ambient*lightColor[0];
	::ambient[1]=f_ambient*lightColor[1];
	::ambient[2]=f_ambient*lightColor[2];
	
	float f_diffuse=param->m_Diffuse;
	::diffuse[0]=f_diffuse*lightColor[0];
	::diffuse[1]=f_diffuse*lightColor[1];
	::diffuse[2]=f_diffuse*lightColor[2];
	
	float f_specular=param->m_Specular;
	::specular[0]=f_specular*lightColor[0];
	::specular[1]=f_specular*lightColor[1];
	::specular[2]=f_specular*lightColor[2];
	
	::specular_power=param->m_SpecularPower;
	
	::spec_threshold=powf(0.001f/max(max(specular[0],specular[1]),specular[2]),1.0f/specular_power);
	
	::light_direction[0]=param->m_LightDirection[0];
	::light_direction[1]=param->m_LightDirection[1];
	::light_direction[2]=param->m_LightDirection[2];
	::light_direction[3]=0.0f;

	::illustrative=param->m_Illustrative;
	::backgroundColor[0]=param->m_BackgroundColor[0];
	::backgroundColor[1]=param->m_BackgroundColor[1];
	::backgroundColor[2]=param->m_BackgroundColor[2];

	return true;
}

