#include "Full/FullRayCaster_basic.core.h"

static inline void GetMinMax(int x,int y,int z,float &minV,float &maxV);
static inline int ConvertToBlockIndex(int voxelIndex,int dimId);
static inline int ConvertToVoxelIndex(int blockIndex,int dimId);

static inline Vector4 CastRay(const Vector4& voxel_zero,const Vector4& voxel_unit,const Vector4& CameraDirection,float tnear,float tfar)
{
	float t1=tnear;

	Vector4 pos1=voxel_zero+voxel_unit*t1;

	float tMaxX,tMaxY,tMaxZ;
	float tDeltaX,tDeltaY,tDeltaZ;

	float t_b;
	
	Vector4 acc(0.0f,0.0f,0.0f,0.0f);
	Vector4 voxel_step=voxel_unit*sampleDistance;

	int x,y,z;
	int stepX,stepY,stepZ;

	if (voxel_unit[0]==0)
		stepX=0;
	else if (voxel_unit[0]<0)
		stepX=-1;
	else
		stepX=1;

	if (voxel_unit[1]==0)
		stepY=0;
	else if (voxel_unit[1]<0)
		stepY=-1;
	else
		stepY=1;

	if (voxel_unit[2]==0)
		stepZ=0;
	else if (voxel_unit[2]<0)
		stepZ=-1;
	else
		stepZ=1;
			
	pos1=voxel_zero+voxel_unit*t1;

	if (voxel_unit[0]==0)
	{
		x=ConvertToBlockIndex(((int)ceilf(pos1[0]-0.5f)-1),0);
		tMaxX=tfar;
	}
	else if (voxel_unit[0]<0)
	{
		x=ConvertToBlockIndex(((int)ceilf(pos1[0]-0.5f)-1),0);
		tMaxX=(ConvertToVoxelIndex(x,0)+0.5f-voxel_zero[0])/voxel_unit[0];
		tDeltaX=-ConvertToVoxelIndex(1,0)/voxel_unit[0];
	}
	else
	{
		x=ConvertToBlockIndex((int)floorf(pos1[0]-0.5f),0);
		tMaxX=(ConvertToVoxelIndex(x+1,0)+0.5f-voxel_zero[0])/voxel_unit[0];
		tDeltaX=ConvertToVoxelIndex(1,0)/voxel_unit[0];
	}

	if (voxel_unit[1]==0)
	{
		y=ConvertToBlockIndex(((int)ceilf(pos1[1]-0.5f)-1),1);
		tMaxY=tfar;
	}
	else if (voxel_unit[1]<0)
	{
		y=ConvertToBlockIndex(((int)ceilf(pos1[1]-0.5f)-1),1);
		tMaxY=(ConvertToVoxelIndex(y,1)+0.5f-voxel_zero[1])/voxel_unit[1];
		tDeltaY=-ConvertToVoxelIndex(1,1)/voxel_unit[1];
	}
	else
	{
		y=ConvertToBlockIndex((int)floorf(pos1[1]-0.5f),1);
		tMaxY=(ConvertToVoxelIndex(y+1,1)+0.5f-voxel_zero[1])/voxel_unit[1];
		tDeltaY=ConvertToVoxelIndex(1,1)/voxel_unit[1];
	}
			
	if (voxel_unit[2]==0)
	{
		z=ConvertToBlockIndex(((int)ceilf(pos1[2]-0.5f)-1),2);
		tMaxZ=tfar;
	}
	else if (voxel_unit[2]<0)
	{
		z=ConvertToBlockIndex(((int)ceilf(pos1[2]-0.5f)-1),2);
		tMaxZ=(ConvertToVoxelIndex(z,2)+0.5f-voxel_zero[2])/voxel_unit[2];
		tDeltaZ=-ConvertToVoxelIndex(1,2)/voxel_unit[2];
	}
	else
	{
		z=ConvertToBlockIndex((int)floorf(pos1[2]-0.5f),2);
		tMaxZ=(ConvertToVoxelIndex(z+1,2)+0.5f-voxel_zero[2])/voxel_unit[2];
		tDeltaZ=ConvertToVoxelIndex(1,2)/voxel_unit[2];
	}

	while (t1<tfar && acc[3]<0.98f)
	{
		t_b=min(min(tMaxX,tMaxY),tMaxZ);
	
		if (t_b>t1+0.001f && BlockEmptyTest(x,y,z))	
		{
			float t=t1;
			Vector4 pos=voxel_zero + voxel_unit * t;		

			while(t<t_b && acc[3]<0.98f)
			{
				Vector4 color=GetShadedColor(pos,CameraDirection);
				float alpha=color[3];
				color[3]=1.0f;
				color*=alpha;		
				acc+=(1.0f-acc[3])*color;				
				t += sampleDistance;
				pos += voxel_step;
			}
		}

		if (tMaxX<tMaxY)
		{
			if (tMaxX<tMaxZ) 
			{
				tMaxX+=tDeltaX;
				x+=stepX;
			}
			else 
			{
				tMaxZ+=tDeltaZ;
				z+=stepZ;
			}
		}
		else
		{
			if (tMaxY<tMaxZ) 
			{
				tMaxY+=tDeltaY;
				y+=stepY;
			}
			else 
			{
				tMaxZ+=tDeltaZ;
				z+=stepZ;
			}
		}
		t1=t_b;
	}
		
	return acc;
}


