#include "Isosurface/IsosurfaceRayCaster_basic.core.h"


static inline Vector4 CastRay(const Vector4& voxel_zero,const Vector4& voxel_unit,const Vector4& CameraDirection,float tnear,float tfar)
{
	float t=tnear;
	Vector4 pos=voxel_zero+voxel_unit*t;
	bool increasing=GetScalarValue(pos)<isovalue;

	float tMaxX,tMaxY,tMaxZ;
	float tDeltaX,tDeltaY,tDeltaZ;
	int x,y,z;
	int stepX,stepY,stepZ;

	if (voxel_unit[0]==0)
	{
		x=floorf(pos[0]);
		tMaxX=tfar;
		stepX=0;
	}
	else if (voxel_unit[0]<0)
	{
		x=ceilf(pos[0])-1;
		tMaxX=(x-voxel_zero[0])/voxel_unit[0];
		tDeltaX=-1.0f/voxel_unit[0];
		stepX=-1;
	}
	else
	{
		x=floorf(pos[0]);
		tMaxX=(x+1-voxel_zero[0])/voxel_unit[0];
		tDeltaX=1.0f/voxel_unit[0];
		stepX=1;
	}

	if (voxel_unit[1]==0)
	{
		y=floorf(pos[1]);
		tMaxY=tfar;
		stepY=0.0f;
	}
	else if (voxel_unit[1]<0)
	{
		y=ceilf(pos[1])-1;
		tMaxY=(y-voxel_zero[1])/voxel_unit[1];
		tDeltaY=-1.0f/voxel_unit[1];
		stepY=-1.0f;
	}
	else
	{
		y=floorf(pos[1]);
		tMaxY=(y+1-voxel_zero[1])/voxel_unit[1];
		tDeltaY=1.0f/voxel_unit[1];
		stepY=1.0f;
	}

	if (voxel_unit[2]==0)
	{
		z=floorf(pos[2]);
		tMaxZ=tfar;
		stepZ=0.0f;
	}
	else if (voxel_unit[2]<0)
	{
		z=ceilf(pos[2])-1;
		tMaxZ=(z-voxel_zero[2])/voxel_unit[2];
		tDeltaZ=-1.0f/voxel_unit[2];
		stepZ=-1.0f;
	}
	else
	{
		z=floorf(pos[2]);
		tMaxZ=(z+1-voxel_zero[2])/voxel_unit[2];
		tDeltaZ=1.0f/voxel_unit[2];
		stepZ=1.0f;
	}

	float s000,s001,s010,s011,s100,s101,s110,s111;
	float t2;
	while (t<tfar)
	{
		t2=min(min(tMaxX,tMaxY),tMaxZ);
		if (t2>tfar) t2=tfar;			
		
		s000=GetScalarValue(x,y,z);
		x+=1;
		s001=GetScalarValue(x,y,z);
		y+=1;
		s011=GetScalarValue(x,y,z);
		x-=1;
		s010=GetScalarValue(x,y,z);
		z+=1;
		s110=GetScalarValue(x,y,z);
		x+=1;
		s111=GetScalarValue(x,y,z);
		y-=1;
		s101=GetScalarValue(x,y,z);
		x-=1;
		s100=GetScalarValue(x,y,z);
		z-=1; 

		if ((s000<isovalue || s001<isovalue || s010<isovalue || s011<isovalue ||
			 s100<isovalue || s101<isovalue || s110<isovalue || s111<isovalue) &&
			(s000>isovalue || s001>isovalue || s010>isovalue || s011>isovalue ||
			 s100>isovalue || s101>isovalue || s110>isovalue || s111>isovalue))
		{
			Vector4 zero2=voxel_zero+voxel_unit*t;

			float u0a=x+1.0f-zero2[0];
			float v0a=y+1.0f-zero2[1];
			float w0a=z+1.0f-zero2[2];

			float u0b=-voxel_unit[0];
			float v0b=-voxel_unit[1];
			float w0b=-voxel_unit[2];

			float u1a=zero2[0]-x;
			float v1a=zero2[1]-y;
			float w1a=zero2[2]-z;

			float u1b=voxel_unit[0];
			float v1b=voxel_unit[1];
			float w1b=voxel_unit[2];		

			float a= u0b*v0b*w0b*s000 + u1b*v0b*w0b*s001 + u0b*v1b*w0b*s010 + u1b*v1b*w0b*s011 +
					 u0b*v0b*w1b*s100 + u1b*v0b*w1b*s101 + u0b*v1b*w1b*s110 + u1b*v1b*w1b*s111;

			float b= (u0a*v0b*w0b+u0b*v0a*w0b+u0b*v0b*w0a)*s000+(u1a*v0b*w0b+u1b*v0a*w0b+u1b*v0b*w0a)*s001 +
					 (u0a*v1b*w0b+u0b*v1a*w0b+u0b*v1b*w0a)*s010+(u1a*v1b*w0b+u1b*v1a*w0b+u1b*v1b*w0a)*s011 +
					 (u0a*v0b*w1b+u0b*v0a*w1b+u0b*v0b*w1a)*s100+(u1a*v0b*w1b+u1b*v0a*w1b+u1b*v0b*w1a)*s101 +
					 (u0a*v1b*w1b+u0b*v1a*w1b+u0b*v1b*w1a)*s110+(u1a*v1b*w1b+u1b*v1a*w1b+u1b*v1b*w1a)*s111;

			float c= (u0b*v0a*w0a+u0a*v0b*w0a+u0a*v0a*w0b)*s000+(u1b*v0a*w0a+u1a*v0b*w0a+u1a*v0a*w0b)*s001 +
					 (u0b*v1a*w0a+u0a*v1b*w0a+u0a*v1a*w0b)*s010+(u1b*v1a*w0a+u1a*v1b*w0a+u1a*v1a*w0b)*s011 +
					 (u0b*v0a*w1a+u0a*v0b*w1a+u0a*v0a*w1b)*s100+(u1b*v0a*w1a+u1a*v0b*w1a+u1a*v0a*w1b)*s101 +
					 (u0b*v1a*w1a+u0a*v1b*w1a+u0a*v1a*w1b)*s110+(u1b*v1a*w1a+u1a*v1b*w1a+u1a*v1a*w1b)*s111;

			float d= u0a*v0a*w0a*s000 + u1a*v0a*w0a*s001 + u0a*v1a*w0a*s010 + u1a*v1a*w0a*s011 +
					 u0a*v0a*w1a*s100 + u1a*v0a*w1a*s101 + u0a*v1a*w1a*s110 + u1a*v1a*w1a*s111 - isovalue;

			float root;
			if (Solve_Cubic_Eq(a,b,c,d,0,t2-t,root))
			{
				pos=zero2+voxel_unit*root;
				return GetColor(pos,CameraDirection,increasing);					
			}
		}

		if (tMaxX<tMaxY)
		{
			if (tMaxX<tMaxZ) 
			{
				tMaxX+=tDeltaX;
				x+=stepX;
			}
			else
			{
				tMaxZ+=tDeltaZ;
				z+=stepZ;
			}
		}
		else
		{
			if (tMaxY<tMaxZ)
			{
				tMaxY+=tDeltaY;
				y+=stepY;
			}
			else
			{
				tMaxZ+=tDeltaZ;
				z+=stepZ;
			}
		}
		t=t2;
	}
	return Vector4(0.0f,0.0f,0.0f,0.0f);
}
/*static inline Vector4 CastRay(const Vector4& voxel_zero,const Vector4& voxel_unit,const Vector4& CameraDirection,float tnear,float tfar)
{
	float t=tnear;
	Vector4 pos=voxel_zero + voxel_unit * t;
	Vector4 voxel_step=voxel_unit*sampleDistance;

	float scalar1=GetScalarValue(pos);
	Vector4 pos1=pos;
	t += sampleDistance;
	pos += voxel_step;
	while(t<tfar)
    {
		float scalar=GetScalarValue(pos);
		if ((scalar>=isovalue && scalar1<=isovalue) || (scalar<=isovalue && scalar1>=isovalue))
		{
			Vector4 isoPos=(isovalue-scalar1)/(scalar-scalar1)*(pos-pos1)+pos1;
			return GetColor(isoPos,CameraDirection,scalar1<=isovalue);
		}
		scalar1=scalar;
		pos1=pos;
		t += sampleDistance;
		pos += voxel_step;
    }
	return Vector4();	

}*/


