#include "MIP/MIPRayCaster_basic.core.h"

static inline void GetMinMax(int x,int y,int z,float &minV,float &maxV);
static inline int ConvertToBlockIndex(int voxelIndex,int dimId);
static inline int ConvertToVoxelIndex(int blockIndex,int dimId);

static inline Vector4 CastRay(const Vector4& voxel_zero,const Vector4& voxel_unit,const Vector4& CameraDirection,float tnear,float tfar)
{
	float t=tnear;
	Vector4 pos=voxel_zero+voxel_unit*t;
	float maxI=GetScalarValue(pos);

	float tMaxX_b,tMaxY_b,tMaxZ_b;
	float tDeltaX_b,tDeltaY_b,tDeltaZ_b;
	float tDeltaX,tDeltaY,tDeltaZ;
	
	int x,y,z;
	int x_b,y_b,z_b;
	int stepX,stepY,stepZ;

	if (voxel_unit[0]==0)
	{
		x_b=ConvertToBlockIndex((int)ceilf(pos[0])-1,0);
		stepX=0;
		tMaxX_b=tfar;
	}
	else if (voxel_unit[0]<0)
	{
		x_b=ConvertToBlockIndex((int)ceilf(pos[0])-1,0);
		stepX=-1;
		tMaxX_b=(ConvertToVoxelIndex(x_b,0)-voxel_zero[0])/voxel_unit[0];
		tDeltaX_b=-ConvertToVoxelIndex(1,0)/voxel_unit[0];
		tDeltaX=-1.0f/voxel_unit[0];
	}
	else
	{
		x_b=ConvertToBlockIndex((int)floorf(pos[0]),0);
		stepX=1;
		tMaxX_b=(ConvertToVoxelIndex(x_b+1,0)-voxel_zero[0])/voxel_unit[0];
		tDeltaX_b=ConvertToVoxelIndex(1,0)/voxel_unit[0];
		tDeltaX=1.0f/voxel_unit[0];
	}

	if (voxel_unit[1]==0)
	{
		y_b=ConvertToBlockIndex((int)ceilf(pos[1])-1,1);
		stepY=0;
		tMaxY_b=tfar;
	}
	else if (voxel_unit[1]<0)
	{
		y_b=ConvertToBlockIndex((int)ceilf(pos[1])-1,1);
		stepY=-1;
		tMaxY_b=(ConvertToVoxelIndex(y_b,1)-voxel_zero[1])/voxel_unit[1];
		tDeltaY_b=-ConvertToVoxelIndex(1,1)/voxel_unit[1];
		tDeltaY=-1.0f/voxel_unit[1];
	}
	else
	{
		y_b=ConvertToBlockIndex((int)floorf(pos[1]),1);
		stepY=1;
		tMaxY_b=(ConvertToVoxelIndex(y_b+1,1)-voxel_zero[1])/voxel_unit[1];
		tDeltaY_b=ConvertToVoxelIndex(1,1)/voxel_unit[1];
		tDeltaY=1.0f/voxel_unit[1];
	}

	if (voxel_unit[2]==0)
	{
		z_b=ConvertToBlockIndex((int)ceilf(pos[2])-1,2);
		stepZ=0;
		tMaxZ_b=tfar;
	}
	else if (voxel_unit[2]<0)
	{
		z_b=ConvertToBlockIndex((int)ceilf(pos[2])-1,2);
		stepZ=-1;
		tMaxZ_b=(ConvertToVoxelIndex(z_b,2)-voxel_zero[2])/voxel_unit[2];
		tDeltaZ_b=-ConvertToVoxelIndex(1,2)/voxel_unit[2];
		tDeltaZ=-1.0f/voxel_unit[2];
	}
	else
	{
		z_b=ConvertToBlockIndex((int)floorf(pos[2]),2);
		stepZ=1;
		tMaxZ_b=(ConvertToVoxelIndex(z_b+1,2)-voxel_zero[2])/voxel_unit[2];
		tDeltaZ_b=ConvertToVoxelIndex(1,2)/voxel_unit[2];
		tDeltaZ=1.0f/voxel_unit[2];
	}

	
	float t_b;
	while (t<tfar)
	{
		float minV,maxV;
		GetMinMax(x_b,y_b,z_b,minV,maxV);
		t_b=min(min(tMaxX_b,tMaxY_b),tMaxZ_b);
		if (tMaxX_b<tMaxY_b)
		{
			if (tMaxX_b<tMaxZ_b) 
			{
				tMaxX_b+=tDeltaX_b;
				x_b+=stepX;
			}
			else 
			{
				tMaxZ_b+=tDeltaZ_b;
				z_b+=stepZ;
			}
		}
		else
		{
			if (tMaxY_b<tMaxZ_b) 
			{
				tMaxY_b+=tDeltaY_b;
				y_b+=stepY;
			}
			else 
			{
				tMaxZ_b+=tDeltaZ_b;
				z_b+=stepZ;
			}
		}
		if ( maxV>maxI)
		{
			if (t_b>tfar) t_b=tfar;

			pos=voxel_zero+voxel_unit*t;
			float tMaxX,tMaxY,tMaxZ;

			if (voxel_unit[0]==0)
			{
				x=floorf(pos[0]);
				tMaxX=tfar;
			}
			else if (voxel_unit[0]<0)
			{
				x=ceilf(pos[0])-1;
				tMaxX=(x-voxel_zero[0])/voxel_unit[0];
			}
			else
			{
				x=floorf(pos[0]);
				tMaxX=(x+1-voxel_zero[0])/voxel_unit[0];
			}

			if (voxel_unit[1]==0)
			{
				y=floorf(pos[1]);
				tMaxY=tfar;
			}
			else if (voxel_unit[1]<0)
			{
				y=ceilf(pos[1])-1;
				tMaxY=(y-voxel_zero[1])/voxel_unit[1];
			}
			else
			{
				y=floorf(pos[1]);
				tMaxY=(y+1-voxel_zero[1])/voxel_unit[1];
			}

			if (voxel_unit[2]==0)
			{
				z=floorf(pos[2]);
				tMaxZ=tfar;
			}
			else if (voxel_unit[2]<0)
			{
				z=ceilf(pos[2])-1;
				tMaxZ=(z-voxel_zero[2])/voxel_unit[2];
			}
			else
			{
				z=floorf(pos[2]);
				tMaxZ=(z+1-voxel_zero[2])/voxel_unit[2];
			}

			float s000,s001,s010,s011,s100,s101,s110,s111;
			float t2;

			while (t<t_b)
			{
				t2=min(min(tMaxX,tMaxY),tMaxZ);
				if (t2>t_b) t2=t_b;			
				
				s000=GetScalarValue(x,y,z);
				x+=1;
				s001=GetScalarValue(x,y,z);
				y+=1;
				s011=GetScalarValue(x,y,z);
				x-=1;
				s010=GetScalarValue(x,y,z);
				z+=1;
				s110=GetScalarValue(x,y,z);
				x+=1;
				s111=GetScalarValue(x,y,z);
				y-=1;
				s101=GetScalarValue(x,y,z);
				x-=1;
				s100=GetScalarValue(x,y,z);
				z-=1; 

				if ((s000>maxI || s001>maxI || s010>maxI || s011>maxI || s100>maxI || s101>maxI || s110>maxI || s111>maxI) )
				{
					Vector4 zero2=voxel_zero+voxel_unit*t;

					float u0a=x+1.0f-zero2[0];
					float v0a=y+1.0f-zero2[1];
					float w0a=z+1.0f-zero2[2];

					float u0b=-voxel_unit[0];
					float v0b=-voxel_unit[1];
					float w0b=-voxel_unit[2];

					float u1a=zero2[0]-x;
					float v1a=zero2[1]-y;
					float w1a=zero2[2]-z;

					float u1b=voxel_unit[0];
					float v1b=voxel_unit[1];
					float w1b=voxel_unit[2];		

					float a= u0b*v0b*w0b*s000 + u1b*v0b*w0b*s001 + u0b*v1b*w0b*s010 + u1b*v1b*w0b*s011 + 
							 u0b*v0b*w1b*s100 + u1b*v0b*w1b*s101 + u0b*v1b*w1b*s110 + u1b*v1b*w1b*s111;

					float b= (u0a*v0b*w0b+u0b*v0a*w0b+u0b*v0b*w0a)*s000+(u1a*v0b*w0b+u1b*v0a*w0b+u1b*v0b*w0a)*s001 +
							 (u0a*v1b*w0b+u0b*v1a*w0b+u0b*v1b*w0a)*s010+(u1a*v1b*w0b+u1b*v1a*w0b+u1b*v1b*w0a)*s011 +
							 (u0a*v0b*w1b+u0b*v0a*w1b+u0b*v0b*w1a)*s100+(u1a*v0b*w1b+u1b*v0a*w1b+u1b*v0b*w1a)*s101 +
							 (u0a*v1b*w1b+u0b*v1a*w1b+u0b*v1b*w1a)*s110+(u1a*v1b*w1b+u1b*v1a*w1b+u1b*v1b*w1a)*s111;

					float c= (u0b*v0a*w0a+u0a*v0b*w0a+u0a*v0a*w0b)*s000+(u1b*v0a*w0a+u1a*v0b*w0a+u1a*v0a*w0b)*s001 +
							 (u0b*v1a*w0a+u0a*v1b*w0a+u0a*v1a*w0b)*s010+(u1b*v1a*w0a+u1a*v1b*w0a+u1a*v1a*w0b)*s011 +
							 (u0b*v0a*w1a+u0a*v0b*w1a+u0a*v0a*w1b)*s100+(u1b*v0a*w1a+u1a*v0b*w1a+u1a*v0a*w1b)*s101 +
							 (u0b*v1a*w1a+u0a*v1b*w1a+u0a*v1a*w1b)*s110+(u1b*v1a*w1a+u1a*v1b*w1a+u1a*v1a*w1b)*s111;

					float d= u0a*v0a*w0a*s000 + u1a*v0a*w0a*s001 + u0a*v1a*w0a*s010 + u1a*v1a*w0a*s011 +
							 u0a*v0a*w1a*s100 + u1a*v0a*w1a*s101 + u0a*v1a*w1a*s110 + u1a*v1a*w1a*s111;

					maxI=max(maxI,GetMax(a,b,c,d,0,t2-t));
				}

				if (tMaxX<tMaxY)
				{
					if (tMaxX<tMaxZ) 
					{
						tMaxX+=tDeltaX;
						x+=stepX;
					}
					else
					{
						tMaxZ+=tDeltaZ;
						z+=stepZ;
					}
				}
				else
				{
					if (tMaxY<tMaxZ)
					{
						tMaxY+=tDeltaY;
						y+=stepY;
					}
					else
					{
						tMaxZ+=tDeltaZ;
						z+=stepZ;
					}
				}
				t=t2;
			}
		}
			
		t=t_b;

	}
	maxI=maxI*mip_scale+mip_shift;
	return Vector4(maxI,maxI,maxI,1.0f);
}
