//#include "ICVolume.h"
#include "CoherentVolume.h"
#include "Vector4.h"

#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

#ifndef min
#define min(a,b)            (((a) < (b)) ? (a) : (b))
#endif


static CoherentVolume *Volume;


//static void* vol_Data;
static int vol_dims[3];
static float vol_spacings[3];
static float vol_size[3];
//static int vol_incs[3];
//static int vol_DataType;

/*template <typename T>
static inline float t_GetScalarValue(T *ptr, const Vector4 &pos)
{
	float a,b,c,d;
	int ipos[3];
	float fpos1,fpos2;

	ipos[0]=(int)pos[0]; ipos[1]=(int)pos[1]; ipos[2]=(int)pos[2];
	ptr+=ipos[0]+ipos[1]*vol_incs[1]+ipos[2]*vol_incs[2];

	fpos2=pos[0]-ipos[0];
	fpos1=1.0f-fpos2;
	
	a=ptr[0]*fpos1+ptr[1]*fpos2;
	b=ptr[vol_incs[1]]*fpos1+ptr[vol_incs[1]+1]*fpos2;
	c=ptr[vol_incs[2]]*fpos1+ptr[vol_incs[2]+1]*fpos2;
	d=ptr[vol_incs[1]+vol_incs[2]]*fpos1+ptr[vol_incs[1]+vol_incs[2]+1]*fpos2;	

	fpos2=pos[1]-ipos[1];
	fpos1=1.0f-fpos2;

	a=a*fpos1+b*fpos2;
	c=c*fpos1+d*fpos2;

	fpos2=pos[2]-ipos[2];
	fpos1=1.0f-fpos2;

	return a*fpos1+c*fpos2;
}
static inline float GetScalarValue(const Vector4& pos)
{
	if (pos[0]<0 || pos[0]>=vol_dims[0]-1 || pos[1]<0 || pos[1]>=vol_dims[1]-1 || pos[2]<0 || pos[2]>=vol_dims[2]-1) return 0.0f;

	switch(vol_DataType)
	{
		case _CHAR:
			return t_GetScalarValue((char*)vol_Data, pos);
		case _UNSIGNED_CHAR:
			return t_GetScalarValue((unsigned char*)vol_Data, pos);
		case _SHORT:
			return t_GetScalarValue((short*)vol_Data, pos);
		case _UNSIGNED_SHORT:
			return t_GetScalarValue((unsigned short*)vol_Data, pos);
		case _INT:
			return t_GetScalarValue((int*)vol_Data, pos);
		case _UNSIGNED_INT:
			return t_GetScalarValue((unsigned int*)vol_Data, pos);
		case _LONG:
			return t_GetScalarValue((long*)vol_Data, pos);
		case _UNSIGNED_LONG:
			return t_GetScalarValue((unsigned long*)vol_Data, pos);
		case _FLOAT:
			return t_GetScalarValue((float*)vol_Data, pos);
		case _DOUBLE:
			return t_GetScalarValue((double*)vol_Data, pos);
		default: return 0.0f;

	}
}*/

static inline float GetScalarValue(int x,int y,int z)
{
	x=max(min(x,vol_dims[0]-1),0);
	y=max(min(y,vol_dims[1]-1),0);
	z=max(min(z,vol_dims[2]-1),0);
	return Volume->GetValue(x,y,z);
}

static inline float GetScalarValue(Vector4 pos)
{
	pos[0]=max(min(pos[0],vol_dims[0]-1),0);
	pos[1]=max(min(pos[1],vol_dims[1]-1),0);
	pos[2]=max(min(pos[2],vol_dims[2]-1),0);
	
	float a,b,c,d;
	int ipos[3];
	float fpos1,fpos2;

	ipos[0]=(int)pos[0]; ipos[1]=(int)pos[1]; ipos[2]=(int)pos[2];
	fpos2=pos[0]-ipos[0];
	fpos1=1.0f-fpos2;
	
	a= GetScalarValue(ipos[0],ipos[1],ipos[2])*fpos1+ GetScalarValue(ipos[0]+1,ipos[1],ipos[2])*fpos2;
	b= GetScalarValue(ipos[0],ipos[1]+1,ipos[2])*fpos1+ GetScalarValue(ipos[0]+1,ipos[1]+1,ipos[2])*fpos2;
	c= GetScalarValue(ipos[0],ipos[1],ipos[2]+1)*fpos1+ GetScalarValue(ipos[0]+1,ipos[1],ipos[2]+1)*fpos2;
	d= GetScalarValue(ipos[0],ipos[1]+1,ipos[2]+1)*fpos1+ GetScalarValue(ipos[0]+1,ipos[1]+1,ipos[2]+1)*fpos2;

	fpos2=pos[1]-ipos[1];
	fpos1=1.0f-fpos2;

	a=a*fpos1+b*fpos2;
	c=c*fpos1+d*fpos2;

	fpos2=pos[2]-ipos[2];
	fpos1=1.0f-fpos2;

	return a*fpos1+c*fpos2;
}

static inline Vector4 GetNormal(const Vector4& pos)
{
	Vector4 curPos;
	Vector4 sample1,sample2;

	curPos=pos;
	curPos[0]-=1.0f;
	sample1[0]=GetScalarValue(curPos);
	curPos=pos;
	curPos[0]+=1.0f;
	sample2[0]=GetScalarValue(curPos);

	curPos=pos;
	curPos[1]-=1.0f;
	sample1[1]=GetScalarValue(curPos);
	curPos=pos;
	curPos[1]+=1.0f;
	sample2[1]=GetScalarValue(curPos);

	curPos=pos;
	curPos[2]-=1.0f;
	sample1[2]=GetScalarValue(curPos);
	curPos=pos;
	curPos[2]+=1.0f;
	sample2[2]=GetScalarValue(curPos);
	
	sample1[3]=0.0;
	sample2[3]=0.0;

	sample2=sample2-sample1;
	sample2[0]/=vol_spacings[0];
	sample2[1]/=vol_spacings[1];
	sample2[2]/=vol_spacings[2];

	sample2.Normalize();

	return sample2;
}

/*static bool PrepareForRendering(ICVolume *volume)
{
	if (!volume) return false;
	vol_DataType=volume->GetDataType();
	volume->GetDimensions(vol_dims);
	volume->GetSpacings(vol_spacings);
	vol_size[0]=vol_dims[0]*vol_spacings[0];
	vol_size[1]=vol_dims[1]*vol_spacings[1];
	vol_size[2]=vol_dims[2]*vol_spacings[2];
	volume->GetIncrements(vol_incs);
	vol_Data=volume->GetData();
	return true;
}*/
static bool SingleScalarVolumePrepare(CoherentVolume *volume)
{
	if (!volume) return false;
	Volume=volume;
	volume->GetDimensions(vol_dims);
	volume->GetSpacings(vol_spacings);
	vol_size[0]=vol_dims[0]*vol_spacings[0];
	vol_size[1]=vol_dims[1]*vol_spacings[1];
	vol_size[2]=vol_dims[2]*vol_spacings[2];
	return true;
}
