#include "VRImage.h"
#include "VRSceneParam.h"
#include <memory.h>

static inline Vector4 CastRay(const Vector4& voxel_zero,const Vector4& voxel_unit,const Vector4& CameraDirection,float tnear,float tfar);

static Matrix4x4 P2MMatrix; 
static Matrix4x4 M2VMatrix; 
static float CroppingBounds[6];	

static Vector4 EyePos;

#ifndef max
#define max(a,b)            (((a) > (b)) ? (a) : (b))
#endif

#ifndef min
#define min(a,b)            (((a) < (b)) ? (a) : (b))
#endif

static inline void ProduceRay(float pixX,float pixY, Vector4& voxel_zero,Vector4& voxel_unit, Vector4& CameraDirection, float& tnear, float& tfar)
{
	Vector4 model_pix=P2MMatrix*Vector4(pixX,pixY,0.0f,1.0f);
	model_pix*=1.0f/model_pix[3];

	CameraDirection=model_pix-EyePos;
	CameraDirection.Normalize();

	voxel_zero=M2VMatrix*EyePos;
	voxel_unit=M2VMatrix*CameraDirection;

	/// Find Start Point
	float t1,t2;

	t1=(CroppingBounds[0]-voxel_zero[0])/voxel_unit[0];
	t2=(CroppingBounds[1]-voxel_zero[0])/voxel_unit[0];

	float tminX=min(t1,t2);
	float tmaxX=max(t1,t2);

	t1=(CroppingBounds[2]-voxel_zero[1])/voxel_unit[1];
	t2=(CroppingBounds[3]-voxel_zero[1])/voxel_unit[1];
	
	float tminY=min(t1,t2);
	float tmaxY=max(t1,t2);

	t1=(CroppingBounds[4]-voxel_zero[2])/voxel_unit[2];
	t2=(CroppingBounds[5]-voxel_zero[2])/voxel_unit[2];

	float tminZ=min(t1,t2);
	float tmaxZ=max(t1,t2);

	tnear=max(max(tminX, tminY), tminZ);
	tnear=max(tnear,0.0f);
	tfar=min(min(tmaxX,tmaxY),tmaxZ);
}

#include "SPMD.h"

static void GenerateImage_MT(int NumberOfThreads, int ThreadID, void* userData)
{
	VRImage *image=(VRImage *)userData;

	int imageSize[2];
	int imageOrigin[2];

	imageSize[0]=image->m_Size[0];
	imageSize[1]=image->m_Size[1];

	imageOrigin[0]=image->m_Origin[0];
	imageOrigin[1]=image->m_Origin[1];

	float* imageData=image->m_Data;
	if (!imageData) return;

	int i,j;
	for (j=imageOrigin[1];j<imageOrigin[1]+imageSize[1];j++)
	{
		for (i=imageOrigin[0]+ThreadID;i<imageOrigin[0]+imageSize[0];i+=NumberOfThreads)
		{
			Vector4 voxel_zero, voxel_unit, CameraDirection;
			float tnear,tfar;
			ProduceRay(i+0.5f,j+0.5f,voxel_zero,voxel_unit,CameraDirection,tnear,tfar);
			Vector4 value(0.0f,0.0f,0.0f,0.0f);
			if (tnear<tfar) value= CastRay(voxel_zero,voxel_unit,CameraDirection,tnear,tfar);
			float *pImage=imageData+(i-imageOrigin[0]+(j-imageOrigin[1])*imageSize[0])*4;
			pImage[0]=max(min(value[0],1.0f),0.0f);
			pImage[1]=max(min(value[1],1.0f),0.0f);
			pImage[2]=max(min(value[2],1.0f),0.0f);
			pImage[3]=max(min(value[3],1.0f),0.0f);
		}
	}
}


static inline void GenerateImage(VRImage *image,
								 const VRSceneParam* sceneParam)
{
	if (!image) return;

	P2MMatrix=sceneParam->m_PixelsToModelMatrix;
	M2VMatrix=sceneParam->m_ModelToVolumeMatrix;

	memcpy(CroppingBounds,sceneParam->m_CroppingBounds,sizeof(float)*6);
	
	EyePos[0]=P2MMatrix[8]/P2MMatrix[11];
	EyePos[1]=P2MMatrix[9]/P2MMatrix[11];
	EyePos[2]=P2MMatrix[10]/P2MMatrix[11];
	EyePos[3]=1.0f;	

	LaunchSPMD(GenerateImage_MT,image);

}
